#if !defined(_THREAD_TEMPLATE_)
#define _THREAD_TEMPLATE_
// listing:	2
// file:	threadtemplate.H
// author:	Allan Kelly

#if !defined(_WINDOWS_)
#include <windows.h>
#endif
#include <cassert>
#include <stdexcept>

// Action to take when dtor is called
enum ThreadTerminateAction
{
        Block,          // halt in dtor until thread terminates
        Kill,           // terminate thread
        Orphan          // orphan thread
};

template<class T, ThreadTerminateAction Action = Block> 
class ThreadTemplate
{
private:
	HANDLE	m_threadHandle;
	DWORD	m_threadId;
	int		m_suspendCount;

public:
	ThreadTemplate() : m_threadHandle(NULL), m_suspendCount(0)
	{
	}

	~ThreadTemplate()
	{
		if (NULL == m_threadHandle)
		{
			return;
		}
		switch (Action)
		{
			case Block :
			{	// prevent destruction until thread has terminated
				if (!IsTerminated())
				{	// if suspended then resume
					while (m_suspendCount > 0)
					{	// may of been suspended more than once
						ResumeThread();
					}

					// has not terminated so wait for it to terminate
					WaitForSingleObject(m_threadHandle, INFINITE);
				}
				break;
			}

			case Kill :
			{	// terminate thread
				if (!IsTerminated())
				{
					int r = TerminateThread(m_threadHandle, 1);
				}
				break;
			}

			case Orphan :
			default :
				// do nothing
				break;
		}  // switch

		// close handle to prevent resource leak
		CloseHandle(m_threadHandle);
	}  // ~ThreadTemplate

	// function must be static so we can take pointer to function
	// return value (unsigned long) is ignored
	// __stdcall enforces calling convention NT expects
	static unsigned long __stdcall StartRun(void *param)
	{	
		T* object = static_cast<T*> (param);	// reclaim object type
		object->Run();
		return 0;
	}

	void RunThread(	T		&x,
					long	stackSize = 0)
	{
		// each object represents one thread
		// therefore, should not launch multiple threads with one object
		assert (NULL == m_threadHandle);

		// other parameters may be dafaulted if needed
		DWORD dwStack = (DWORD) stackSize;

		// create a  thread
		m_threadHandle = CreateThread(  
								NULL,			// pointer to thread security attributes  
								dwStack,		// initial thread stack size, in bytes
								StartRun,		// pointer to thread function  
								&x,				// argument for new thread 
								0,				// creation flags 
								&m_threadId);	// pointer to returned thread identifier 
		DWORD last = GetLastError();	// useful for debug stepping

		if (NULL == m_threadHandle)
		{
			throw std::runtime_error("Failed to start thread");
		}
#if defined(SINGLE_THREAD_DEBUGGING)
		WaitForSingleObject(m_threadHandle, INFINITE);
#endif
	}

	bool IsTerminated()
	{
		if (NULL == m_threadHandle)
		{	// not run yet so is terminated
			return true;
		}

		DWORD w = WaitForSingleObject(m_threadHandle, 0);
		if (WAIT_OBJECT_0 == w)
		{
			return true;
		}

		return false;
	}

	bool IsSuspended()
	{
		return m_suspendCount > 0;
	}

	void SuspendThread()
	{
		m_suspendCount = ::SuspendThread(m_threadHandle);
	}

	void ResumeThread()
	{
		m_suspendCount = ::ResumeThread(m_threadHandle);
	}

	void Complete()
	{
		if (!IsTerminated())
		{
			while (m_suspendCount > 0)
			{
				ResumeThread();
			}

			WaitForSingleObject(m_threadHandle, INFINITE);
		}
	}
};  // ThreadTemplate
#endif
